#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#include <mef.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

int main(int argc, char **argv) {
  /* Some parameters */
  unsigned int s = 3;    /* mixtures */
  unsigned int k = 3;    /* compnents */
  unsigned int n = 5000; /* observation by mixtures */
  unsigned int d = 1;    /* dimension */

  /* Random generator initialization */
  const gsl_rng_type *T;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  gsl_rng *rng = gsl_rng_alloc(T);

  /* Mixtures creation */
  Gaussian_family* ef = Gaussian();
  mixture** mixtures = malloc(s * sizeof(mixture*));
  for (unsigned int i=0; i<s; i++) {
    mixtures[i] = mixture_create(k);
    for (unsigned int j=0; j<k; j++) {
      mixtures[i]->ef[j] = (family*)ef;
    }
  }

  /* Parameters of the mixtures */
  mixtures[0]->params[0]  = (param*)Gaussian_create_source(ef,   0., 4.);
  mixtures[0]->params[1]  = (param*)Gaussian_create_source(ef, -10., 3.);
  mixtures[0]->params[2]  = (param*)Gaussian_create_source(ef,   5., 5.);
  mixtures[0]->weights[0] = 0.2;
  mixtures[0]->weights[1] = 0.5;
  mixtures[0]->weights[2] = 0.3;

  mixtures[1]->params[0]  = (param*)Gaussian_create_source(ef,   0., 3.);
  mixtures[1]->params[1]  = (param*)Gaussian_create_source(ef,  -8., 3.);
  mixtures[1]->params[2]  = (param*)Gaussian_create_source(ef,   5., 5.);
  mixtures[1]->weights[0] = 0.1;
  mixtures[1]->weights[1] = 0.4;
  mixtures[1]->weights[2] = 0.5;

  mixtures[2]->params[0]  = (param*)Gaussian_create_source(ef,   0., 3.);
  mixtures[2]->params[1]  = (param*)Gaussian_create_source(ef,  -8., 3.);
  mixtures[2]->params[2]  = (param*)Gaussian_create_source(ef,   5., 5.);
  mixtures[2]->weights[0] = 0.3;
  mixtures[2]->weights[1] = 0.3;
  mixtures[2]->weights[2] = 0.4;

  for (unsigned int i=0; i<s; i++) {
    printf("True mixture %i:\n", i);
    mixture_fprint(stdout, mixtures[i], SOURCE);
  }

  /* Generation of samples */
  double      **data  = malloc(s * sizeof(double*));
  unsigned int *sizes = malloc(s * sizeof(unsigned int));
  for (unsigned int i=0; i<s; i++) {
    data[i]  = malloc(n * sizeof(double));
    sizes[i] = n;
    for(unsigned int j=0; j<n; j++) {
      data[i][j] = *mixture_rand(mixtures[i], rng);
    }
  }

  /* coEM */
  coem* coem1 = coem_create((family*)ef, k, s, data, sizes);
  coem1->maxiter  = 100;
  coem1->treshold = 1e-5;

  /* Initialization of coEM with kmeans */
  double *data_all = malloc(s * n * sizeof(double));
  unsigned int i=0;
  for (unsigned int l=0; l<s; l++) {
    for (unsigned int j=0; j < n; j++) {
      data_all[i] = data[l][j];
      i++;
    }
  }

  double best_error = DBL_MAX;
  kmeans* km_tmp;
  kmeans* km;
  for (unsigned int i=0; i<10; i++) {
    km_tmp = kmeans_create(k, data_all, s*n, d);
    kmeans_initialize_random(km_tmp, rng);
    kmeans_run(km_tmp);

    if (km_tmp->old_error < best_error) {
      printf("%f\n", km_tmp->old_error);
      km = km_tmp;
        best_error = km_tmp->old_error;
    }
    else {
      kmeans_destroy(km_tmp);
    }
  }
  for (int i=0; i<k; i++){
    printf("weights[%i]=%f\n", i, km->weights[i]);
  }
  unsigned int **coaffectation = malloc(s * sizeof(unsigned int*));
  double       **coweights     = malloc(s * sizeof(double*));
  for (unsigned int l=0; l<s; l++) {
    coaffectation[l] = &km->affectation[l*n];
    coweights[l]     = km->weights;
  }

  printf("Initialization:\n");
  coem_initialize_from_coclustering(coem1, coweights, coaffectation);
  for (unsigned int i=0; i<s; i++) {
    printf("Mixture %i:\n", i);
    mixture_fprint(stdout, coem1->mixtures[i], SOURCE);
  }
  printf("\nRunning:\n");
  mixture** mix = coem_run(coem1);
  printf("Number of iteration: %i\n", coem1->iter);
  for (unsigned int i=0; i<s; i++) {
    printf("Mixture %i:\n", i);
    mixture_fprint(stdout, mix[i], SOURCE);
  }

  /* Another version, printing information at each step */
  printf("\nStep by step:\n");
  coem* coem2 = coem_create((family*)ef, k, s, data, sizes);
  coem2->treshold = 1e-5;
  coem_initialize_from_coclustering(coem2, coweights, coaffectation);
  /* we make the necessary number of steps to get convergence, as found
     with the first coem */
  for (unsigned int i=0; i<coem1->iter; i++) {
    coem_step(coem2);
    printf("Step %i, ll=%f\n", coem2->iter, coem2->ll);
  }
  for (unsigned int i=0; i<s; i++) {
    printf("Mixture %i:\n", i);
    mixture_fprint(stdout, coem2->mixtures[i], SOURCE);
  }

  /* Some cleaning */
  kmeans_destroy(km);
  coem_destroy(coem1);
  coem_destroy(coem2);
}

